import urllib.parse

class URLFactory:
    def __init__(self, base):
        self.base = base

    def make_webapi(self, page, args, format="json"):
        query = '&'.join([
            f'{k}={urllib.parse.quote(v)}' for k,v in args.items()
        ])
        return f"{self.base}/{page}.php?format={format}&{query}"

    def make_clientapi(self, page):
        return f"http://localhost:3490/v1/{page}"



import aiohttp
import json
import os
from .api.url_factory import URLFactory
from .structures.pulse import *
from .structures.stats import *
from .structures.teams import *
from .structures.users import *


class Client:
    """
        Client for interacting with WhatPulse's web API and [client API].
        The client always assumes the local WhatPulse client has it's client API
        enabled on port 3490. Customizable ports are not supported.
    """
    def __init__(self):
        self._url = URLFactory("http://api.whatpulse.org")
        self._session = aiohttp.ClientSession()
        
    async def _check_avil(self) -> bool:
        try:
            async with self._session.get("http://localhost:3490") as r:
                return r.status == 200
        except aiohttp.ClientConnectionError as e:
            return False

    async def _get(self, url) -> dict:
        try:
            async with self._session.get(url) as r:
                if r.status == 404:
                    return None
                elif r.status != 200:
                    raise RuntimeError(
                        "from Client: " f"Server returned non-200 code: {r.status}"
                    )
                elif r.content_length == 0:
                    return None
                j = json.loads(await r.text())
                if "err" in j:
                    raise RuntimeError(f"from Client: {j['err']}")

                return j
        except aiohttp.ClientConnectionError as e:
            return None

    async def get_user(self, id) -> User:
        """
            Retrieve a WhatPulse user from the web API.
        """
        url = self._url.make_webapi("user", {"user": id})
        j = await self._get(url)
        if j is None:
            return None

        return User(j)

    async def get_team(self, id) -> Team:
        """
            Retrieve a WhatPulse team from the web API.
        """
        url = self._url.make_webapi("team", {"team": id})
        j = await self._get(url)
        if j is None:
            return None

        return Team(j)

    async def get_user_pulses(self, id) -> list:
        """
            Retrieve a list of a WhatPulse user's pulses from the web API.
        """
        url = self._url.make_webapi("pulses", {"user": id})
        j = await self._get(url)
        if j is None:
            return None

        return [Pulse(j[k]) for k in j]

    async def get_team_pulses(self, id) -> list:
        """
            Retrieve a list of a WhatPulse team's pulses from the web API.
        """
        url = self._url.make_webapi("pulses", {"team": id})
        j = await self._get(url)
        if j is None:
            return []

        return [Pulse(j[k]) for k in j]

    async def get_local_totals(self) -> tuple:
        """
            Retrieve the account totals from the local WhatPulse client.
            Returns a tuple containing (Stats, Ranks).
        """
        url = self._url.make_clientapi("account-totals")
        j = await self._get(url)
        if j is None:
            return None

        return Stats(j), Ranks(j["ranks"])

    async def pulse_local(self) -> None:
        """
            Executes a pulse on the local WhatPulse client.
        """
        url = self._url.make_clientapi("pulse")

        async with self._session.post(url) as r:
            if r.status != 200:
                raise RuntimeError(
                    "from Client: " f"Server returned non-200 code: {r.status}"
                )

    async def get_local_unpulsed(self) -> Stats:
        """
            Retrieve the unpulsed stats from the local WhatPulse client.
        """
        url = self._url.make_clientapi("unpulsed")
        j = await self._get(url)
        if j is None:
            return None

        return Stats(j)

    async def close(self):
        await self._session.close()



from datetime import datetime
from .stats import Stats, Ranks


class Computer:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v

        self._src = src

        self.id = int(src["computerid"])
        self.name = src["name"]
        self.stats = Stats(src)

        if src["lastpulseunixtimestamp"] is not None:
            self.last_pulse = datetime.utcfromtimestamp(
                int(src["lastpulseunixtimestamp"])
            )
        else:
            self.last_pulse = datetime.utcfromtimestamp(0)


class User:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v
        
        self._src = src

        self.id = int(src["userid"])
        self.name = src["accountname"]
        self.country = src["country"]
        self.country_short = src["tld"]
        self.homepage = src["homepage"]
        self.team_id = int(src["team"]["TeamID"])

        self.joined = datetime.utcfromtimestamp(int(src["datejoinedunixtimestamp"]))
        self.last_pulse = datetime.utcfromtimestamp(int(src["lastpulseunixtimestamp"]))

        self.stats = Stats(src)
        self.ranks = Ranks(src["ranks"])
        self.computers = [Computer(s) for s in src["computers"].values()]



from datetime import datetime
from .stats import Stats, Ranks


class Team:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v
        
        self._src = src

        self.id = src["teamid"]
        self.name = src["name"]
        self.description = src["description"]
        self.users = src["users"]

        self.stats = Stats(src)
        self.ranks = Ranks(src["ranks"])

        self.formed = datetime.utcfromtimestamp(int(src["dateformedunixtimestamp"]))
        self.founder = src["founder"]

        # the docs lied to me :(
        # self.subteams = [Subteam(s) for s in src["subteams"].values()]


class Subteam:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v
        
        self._src = src

        self.id = int(src["subteamid"])
        self.name = src["subteamname"]
        self.founder = src["founder"]
        self.formed = datetime.utcfromtimestamp(int(src["dateformedynixtimestamp"]))
        self.users = int(src["users"])

        self.stats = Stats(src)
        self.ranks = Ranks(src["ranks"])



from datetime import datetime
from .stats import Stats


class Pulse:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v

        self._src = src
        
        self.user_id = int(src["userid"])
        self.user_name = src["username"]
        self.computer_name = src["computer"]
        self.os_name = src["os"]

        self.stats = Stats(src)

        self.timestamp = datetime.utcfromtimestamp(int(src["timestamp"]))



import asyncio
import pulseapi

CLIENT = pulseapi.Client()

async def __userstats(*args):
    """Returns the stats of a user.
Usage: user <id or name>"""
    if len(args) > 0:
        user = await CLIENT.get_user(' '.join(args))
        if user is None:
            return "Could not fetch user."
        else:
            return f"""Clicks - {user.stats.clicks}
Keys - {user.stats.keys}
Download - {user.stats.download}
Upload - {user.stats.upload}
Uptime - {user.stats.uptime}"""
    else:
        return "Please specify user. (ID or name)"

async def __teamstats(*args):
    """Returns the stats of a team.
Usage: team <id or name>"""
    if len(args) > 0:
        team = await CLIENT.get_team(' '.join(args))
        if team is None:
            return "Could not fetch user."
        else:
            return f"""Clicks - {team.stats.clicks}
Keys - {team.stats.keys}
Download - {team.stats.download}
Upload - {team.stats.upload}
Uptime - {team.stats.uptime}"""
    else:
        return "Please specify team. (ID or name)"

async def __pulse(*args):
    """Executes a pulse on the local client."""
    await CLIENT.pulse_local()

async def __quit(*args):
    """Exit the CLI."""
    pass

# ==================== PULSEAPI ENDS HERE ==================== #

async def _help(*args):
    """Gets help on a command.
Usage: help <command>"""
    if len(args) > 0:
        if args[0].lower() in CMD_LIST:
            return CMD_LIST[args[0].lower()].__doc__
        else:
            return f"ERROR: Command `{args[0]}` not found."
    else:
        return "Type the name of the command after `help` to get help on that"
        "command."

async def __commands(*args):
    """Returns a list of commands."""
    return ('\n'.join([
            f"{n} - {c.__doc__.splitlines()[0]}"  for n,c in CMD_LIST.items()
        ]))

CMD_LIST = {
    "help":_help,
    "cmd":__commands,
    "user":__userstats,
    "team":__teamstats,
    "pulse":__pulse,
    "quit":__quit
}

async def main():
    print(f"""puleapi v{pulseapi.__version__}
Type `help` for help.
Type `cmd` for a list of commands.
""")
    while True:
        raw = input("? ")
        cmd = raw.lower().split(" ")[0]
        args = raw.split(" ")[1:]

        if cmd == "quit":
            await CLIENT.close()
            break
        elif cmd in CMD_LIST:
            print(await CMD_LIST[cmd](*args))
        else:
            print(f"ERROR: Invalid command `{cmd}`.")
        print()

if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())


def get_uptime_tuple(sec):
    days, t = divmod(sec, 24 * 60 ** 2)
    hours, t = divmod(t, 60 ** 2)
    minutes, t = divmod(t, 60)
    seconds = t
    return days, hours, minutes, seconds


class Stats:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower()] = v

        self._src = src

        self.keys = int(src["keys"])
        self.clicks = int(src["clicks"])

        if "downloadmb" in src:
            self.download = int(src["downloadmb"])
        else:
            self.download = int(src["download"])

        if "uploadmb" in src:
            self.upload = int(src["uploadmb"])
        else:

            self.upload = int(src["upload"])
        if "uptime" in src:
            uptime = int(src["uptime"])
        else:
            uptime = int(src["uptimeseconds"])
        self.uptime = get_uptime_tuple(uptime)

        if "pulses" in src:
            self.pulses = int(src["pulses"])
        else:
            self.pulses = -1


class Ranks:
    def __init__(self, src):
        _src = src.copy()
        src = {}
        for k,v in _src.items():
            src[k.lower().replace("rank_", "")] = v

        self._src = src

        self.keys = int(src["keys"])
        self.clicks = int(src["clicks"])
        self.download = int(src["download"])
        self.upload = int(src["upload"])
        self.uptime = int(src["uptime"])



import setuptools
from pulseapi import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pulseapi",
    version=__version__,
    author="qeaml",
    author_email="qeaml@wp.pl",
    description="qeaml's wrapper for WhatPulse's API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/QeaML/pulseapi",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	license='MIT',
    python_requires='>=3.6',
	install_requires=['aiohttp'],
)